"""conversion functions across different formats.
"""
import logging
from typing import Any
from typing import Dict
from typing import List
from typing import Tuple

import h5py
import numpy as np

from ._iterator_access import IteratorAccess
from .hdf5 import hdf5append


logger = logging.getLogger(__name__)


# TODO this file should be abstracted layer of each fileformat R/W.
# Move this function to hdf5
def convert_iterator_access_to_hdf5(
    iterator_access: IteratorAccess,
    hdf5file: str,
    key_pairs: Dict[str, str],
    image_keys: List[str] = [],
) -> int:
    """Convert IteratorAccess data format into HDF5 format.

    Args:
        iterator_access (IteratorAccess): Iterator access format. Each iteration should return dictionary.
        hdf5file (str): file path for the HDF5 file.
        key_pairs (dict): {data_key: hdf5_key}
        image_keys (list of str): data_key list for specifying image data (more than 1 dimensional).
            Used to define the block size of hdf5.
    Returns:
        int ... number of data processed.
    """
    logger.info(f"Convert data {iterator_access.FORMAT} from {iterator_access.file_name}")
    logger.info(f"The key-value map (Iterator data -> Hdf5 key): {key_pairs}")
    logger.info(f"Saving events into {hdf5file}")
    # TODO add filecheck of the hdf5.
    i = 0
    with h5py.File(hdf5file, "w") as f:
        for iter_data in iterator_access:
            if i == 0:
                for (k, v) in key_pairs.items():
                    if k in image_keys:
                        image = iter_data[k]
                        maxshape = (None,) + image.shape[1:]
                        f.create_dataset(
                            v,
                            data=image,
                            maxshape=maxshape,
                            compression="gzip",
                            compression_opts=9,
                        )
                    else:
                        f.create_dataset(
                            v,
                            data=iter_data[k],
                            maxshape=(None,),
                            compression="gzip",
                            compression_opts=9,
                        )
            else:
                for (k, v) in key_pairs.items():
                    hdf5append(f[v], iter_data[k])
            i += len(iter_data)
    logger.info(f"Done. Total {i} data points are processed.")
    return i


# TODO this file should be abstracted layer of each fileformat R/W.
# Move this function to text
def write_to_text(
    event: np.ndarray,
    file_name: str
) -> None:
    np.savetxt(file_name, event[:, [2, 1, 0, 3]], fmt=['%.9f', '%d', '%d', '%d'])



"""Event Vision Library.
=====
Provides
  1. Flexible and user-friendly interface of event data.
  2. Read/write different format and types of event data.
  3. Data loader for existing research datasets.
  4. Preprocessing.
  5. Algorithms.
  6. Visualization.
"""

from .version import __version__, git_revision as __git_version__  # type: ignore



from typing import Optional, List, Tuple

import torch

from ..base_model import BaseModel
from ..model.unet import UNetRecurrent


class BaseE2VID(BaseModel):
    def __init__(self, config: dict) -> None:
        super().__init__(config)

        assert('num_bins' in config)
        self.num_bins = int(config['num_bins'])  # number of bins in the voxel grid event tensor

        try:
            self.skip_type = str(config['skip_type'])
        except KeyError:
            self.skip_type = 'sum'

        try:
            self.num_encoders = int(config['num_encoders'])
        except KeyError:
            self.num_encoders = 4

        try:
            self.base_num_channels = int(config['base_num_channels'])
        except KeyError:
            self.base_num_channels = 32

        try:
            self.num_residual_blocks = int(config['num_residual_blocks'])
        except KeyError:
            self.num_residual_blocks = 2

        try:
            self.norm = str(config['norm'])
        except KeyError:
            self.norm = None  # type: ignore

        try:
            self.use_upsample_conv = bool(config['use_upsample_conv'])
        except KeyError:
            self.use_upsample_conv = True


class E2VIDRecurrent(BaseE2VID):
    """
    Recurrent, UNet-like architecture where each encoder is followed by a ConvLSTM or ConvGRU.
    """

    def __init__(self, config: dict) -> None:
        super(E2VIDRecurrent, self).__init__(config)

        try:
            self.recurrent_block_type = str(config['recurrent_block_type'])
        except KeyError:
            self.recurrent_block_type = 'convlstm'  # or 'convgru'

        self.unetrecurrent = UNetRecurrent(num_input_channels=self.num_bins,
                                           num_output_channels=1,
                                           skip_type=self.skip_type,
                                           recurrent_block_type=self.recurrent_block_type,
                                           activation='sigmoid',
                                           num_encoders=self.num_encoders,
                                           base_num_channels=self.base_num_channels,
                                           num_residual_blocks=self.num_residual_blocks,
                                           norm=self.norm,
                                           use_upsample_conv=self.use_upsample_conv)

    def forward(self, event_tensor: torch.Tensor,  # type: ignore
                prev_states: Optional[List[torch.Tensor]]) -> Tuple[torch.Tensor,
                                                                    List[torch.Tensor]]:
        """
        :param event_tensor: N x num_bins x H x W
        :param prev_states: previous ConvLSTM state for each encoder module
        :return: reconstructed image, taking values in [0,1].
        """
        img_pred, states = self.unetrecurrent.forward(event_tensor, prev_states)
        return img_pred, states



"""Diffenet file formats for various event cameras and datasets.
"""
from .aedat import IteratorAedat4Event
from .aedat import IteratorAedat4Frame
from .aedat import IteratorAedat4Imu
from .aedat import IteratorAedat4Trigger
from .conversion import convert_iterator_access_to_hdf5
from .evk import IteratorEvk3
from .hdf5 import load_event_timestamp_hdf5
from .hdf5 import load_hdf5
from .hdf5 import open_hdf5
from .text import IteratorTextEvent
from .text import IteratorTextFrame
from .text import IteratorTextImu
from .text import IteratorTextPose



"""Voxel-grid representation.
"""
from typing import Any
from typing import Tuple

import numpy as np


class VoxelGrid:
    """Create a voxel grid from events.
    Implementation inspired by https://github.com/uzh-rpg/rpg_e2vid.

    Args:
        image_shape: (height, width)
        num_bins: number of bins in the temporal axis of the voxel grid
    """
    def __init__(self, image_shape: Tuple[int, int], num_bins: int) -> None:
        assert image_shape[0] > 0
        assert image_shape[1] > 0
        assert num_bins > 0
        self.image_shape = image_shape
        self.num_bins = num_bins

    def __call__(self, events: np.ndarray) -> np.ndarray:
        """Create voxel grid.

        Args:
            events: events: a NumPy array of size [n x d], where n is the number of events and d = 4.
                    Every event is encoded with 4 values (y, x, t, p).

        Returns:
            A voxelized representation of the event data.
        """
        assert events.shape[1] == 4
        height, width = self.image_shape
        voxel_grid = np.zeros((self.num_bins, height, width), np.float32).ravel()
        y = events[:, 0].astype(int)
        x = events[:, 1].astype(int)
        t = events[:, 2]
        p = events[:, 3]
        p[p == 0] = -1

        # Normalize the event timestamps so that they lie between 0 and num_bins
        t_min, t_max = np.amin(t), np.amax(t)
        delta_t = t_max - t_min

        if delta_t == 0:
            delta_t = 1

        t = (self.num_bins - 1) * (t - t_min) / delta_t
        tis = t.astype(int)
        dts = t - tis
        vals_left = p * (1.0 - dts)
        vals_right = p * dts
        valid_indices = tis < self.num_bins

        np.add.at(
            voxel_grid,
            x[valid_indices] + y[valid_indices] * width + tis[valid_indices] * width * height,
            vals_left[valid_indices],
        )

        valid_indices = (tis + 1) < self.num_bins
        np.add.at(
            voxel_grid,
            x[valid_indices] + y[valid_indices] * width + (tis[valid_indices] + 1) * width * height,
            vals_right[valid_indices],
        )

        voxel_grid = np.reshape(voxel_grid, (self.num_bins, height, width))

        return voxel_grid



import os
import logging
import sys
from typing import Tuple

import numpy as np
import urllib.request

from .e2vid_module.utils.loading_utils import load_model, get_device
from .e2vid_module.image_reconstructor import ImageReconstructor
from ...representation import VoxelGrid  # TODO: import correctly from evlib

logging.basicConfig(level=logging.INFO, format='%(asctime)s [%(levelname)s] %(message)s')
logger = logging.getLogger('ev_lib')

MODEL_PATH = os.path.join(os.path.dirname(os.path.realpath(__file__)),
                                  "..", "..", "..", "..", "..", "artifacts",
                                  "E2VID_lightweight.pth.tar")


def progress_bar(count: int, block_size: int, total_size: int) -> None:
    completed = count * block_size
    progress = completed / total_size
    bar_length = 50
    filled_length = int(progress * bar_length)
    bar = "=" * filled_length + "-" * (bar_length - filled_length)
    sys.stdout.write(f"\rDownloading: [{bar}] {progress:.2%}")
    sys.stdout.flush()


class DictionatyPropagation():
    def __init__(self, *initial_data, **kwargs):  # type: ignore
        for dictionary in initial_data:
            for key in dictionary:
                setattr(self, key, dictionary[key])
        for key in kwargs:
            setattr(self, key, kwargs[key])


class E2Vid:
    """Integration of e2vid image reconstruction based on
    https://github.com/uzh-rpg/rpg_vid2e
    
    This wrapper allows you to instantiate the model and
    reconstruct images by using __call__.
    Note that you need to iteratively call this module with subsequent
    event batches for good reconstruction results.
    The number of events per batch can affect the reconstruction results.

    Args:
        image_shape: (height, width)
        use_gpu: if GPU should be used for image reconstruction
    """
    def __init__(self,
                 image_shape: Tuple[int, int],
                 use_gpu: bool = True,
                 unsharp_mask_amount: float = 0.3,
                 unsharp_mask_sigma: float = 1.0,
                 bilateral_filter_sigma: float = 0.0,
                 flip: bool = False,
                 Imin: float = 0.0,
                 Imax: float = 1.0,
                 auto_hdr: bool = False,
                 auto_hdr_median_filter_size: int = 10,
                 no_normalize: bool = False,
                 no_recurrent: bool = False,                 
                 ) -> None:
        assert image_shape[0] > 0
        assert image_shape[1] > 0
        self.image_shape = image_shape
        
        config_dict = {
            "use_gpu": use_gpu,
            "unsharp_mask_amount": unsharp_mask_amount,
            "unsharp_mask_sigma": unsharp_mask_sigma,
            "bilateral_filter_sigma": bilateral_filter_sigma,
            "flip": flip,
            "Imin": Imin,
            "Imax": Imax,
            "auto_hdr": auto_hdr,
            "auto_hdr_median_filter_size": auto_hdr_median_filter_size,
            "no_normalize": no_normalize,
            "no_recurrent": no_recurrent,
        }
        config = DictionatyPropagation(config_dict)

        self._download_model()
        self.model = load_model(MODEL_PATH)
        self.device = get_device(use_gpu)

        self.model = self.model.to(self.device)
        self.model.eval()
        n_bins: int = self.model.num_bins
        self.voxelizer = VoxelGrid(image_shape, n_bins)
        self.reconstructor = ImageReconstructor(self.model, image_shape[0],
                                                image_shape[1], n_bins,
                                                config)

    def _download_model(self) -> None:
        e2vid_pretrained_url = "http://rpg.ifi.uzh.ch/data/E2VID/models/E2VID_lightweight.pth.tar"
        
        if not os.path.exists(MODEL_PATH):
            os.makedirs(os.path.dirname(MODEL_PATH), exist_ok=True)
            logger.info(f"Downloading pretrained model file.")
            urllib.request.urlretrieve(e2vid_pretrained_url, filename=MODEL_PATH,
                                       reporthook=progress_bar)
            logger.info("Done.")

    def __call__(self, events:np.ndarray) -> np.ndarray:
        """Reconstruct image from event batch

        Args:
            events: a NumPy array of size [n x d], where n is the number of events and d = 4.
                    Every event is encoded with 4 values (y, x, t, p).

        Returns:
            The reconstructed image of size <image_shape>.
        """
        assert events.shape[1] == 4
        event_tensor = self.voxelizer(events)
        image = self.reconstructor.update_reconstruction(event_tensor)
        return image


"""hdf5 formats for various existing datasets.
"""
import logging
from typing import Any
from typing import Dict
from typing import List
from typing import Tuple
from typing import Union

import h5py
import numpy as np


H5_SET = Tuple[str, str, Any]

logger = logging.getLogger(__name__)


def load_hdf5(
    path: str, key_dtype_pairs: Union[List[H5_SET], H5_SET]
) -> Dict[str, np.ndarray]:
    """Load all the contents of .hdf5 file at once.
    
    Args:
        path (str): Path to the .hdf5 file.
        key_dtype_pairs (list of tuple): A triplet, or list of triplets of \\
            (key of the return dictionary, key of the hdf5 file data, data type for numpy).\\
            For example,
            [("ts", "raw_events/ts", np.int32), ("x", "raw_events/x", np.int16), ...]

    Returns:
        dict: {key of the return dictionary: np.ndarray}
    """
    if isinstance(key_dtype_pairs, tuple):
        key_dtype_pairs = [key_dtype_pairs]  # make it list.
    f = h5py.File(path, "r")
    data = {k: np.array(f[v], dtype=t) for (k, v, t) in key_dtype_pairs}
    f.close()
    return data


def open_hdf5(path: str) -> Any:
    """Open .hdf5 file, not to load them at once.
    
    Args:
        path (str): Path to the .hdf5 file.

    Returns:
        (Any) opened hdf5 object.
    """
    return h5py.File(path, "r")


def load_event_timestamp_hdf5(
    path: str, key_pairs: Tuple[str, str], dtype: type = np.int32
) -> Dict[str, np.ndarray]:
    """For utility: load only timestamps from the hdf5 data.
    
    Args:
        path (str): Path to the .hdf5 file.
        key_pairs : The tuple of \\
            (key of the return dictionary, key of the hdf5 file data)
    
    Returns:
        dict
    """
    data = load_hdf5(path, list(key_pairs + (dtype,)))  # type: ignore
    if np.max(data[key_pairs[0]]) == np.iinfo(dtype):
        w = f"Please check the size of the data and data type. {dtype} may not be enough."
        logger.warning(w)
    return data


def hdf5append(data: Any, new_arr: np.ndarray) -> None:
    """For utility: append an array to hdf5 data.

    Args:
        data (Any): _description_
        new_arr (np.ndarray): _description_
    """
    n_old = data.shape[0]
    n_new = n_old + len(new_arr)
    data.resize(n_new, axis=0)
    data[n_old:] = new_arr



"""Data type for multiple events."""
import copy
from dataclasses import dataclass
from typing import Any, List, Optional, Union

import numpy as np
import numpy.typing as npt

from .raw_event import RawEvent


@dataclass
class RawEvents:
    """Dataclass for a batch of raw events."""
    x: npt.NDArray[np.int16]  # [0, width]
    y: npt.NDArray[np.int16]  # [0, height]
    timestamp: npt.NDArray[np.float64]
    polarity: npt.NDArray[np.bool_]   # true for positive, false for negative

    # Shortcuts
    @property
    def p(self) -> npt.NDArray[np.bool_]:
        """Alias for polatiry.
        """
        return self.polarity

    @property
    def t(self) -> npt.NDArray[np.float64]:
        """Alias for timestamp.
        """
        return self.timestamp

    # Build-ins
    def __getitem__(self, index: Union[int, str]) -> npt.NDArray[np.float64]:
        """Get an event as numpy array.

        Args:
            index (int): index in the batch

        Returns:
            npt.NDArray[np.float64]: 1-d numpy array.
        """
        if isinstance(index, str):
            return getattr(self, index)  # type: ignore
        return np.array([self.y[index], self.x[index], self.t[index], self.p[index]], dtype=np.float64)

    def __len__(self) -> int:
        """Get the number of events.

        Returns:
            int: n_events
        """
        return len(self.x)

    @property
    def n(self) -> int:
        """Alias for len().
        """
        return len(self)

    # Utility
    def append(self, e: RawEvent) -> None:
        """Append one event to the event batch object.

        Args:
            e (RawEvent): Event to be appended.
        """
        self.x = np.append(self.x, e.x)
        self.y = np.append(self.y, e.y)
        self.timestamp = np.append(self.timestamp, e.timestamp)
        self.polarity = np.append(self.x, e.x)

    def as_numpy(self) -> npt.NDArray[np.float64]:
        """Convert event object into 2-d numpy array.

        Returns:
            npt.NDArray[np.float64]: 2-d numpy array, [n_events, 4].
        """
        return np.stack([self.y, self.x, self.t, self.p]).astype(np.float64).T



"""Histogram representation.
"""
from typing import Any
from typing import Tuple

import numpy as np


class Histogram:
    """Create a 2D histogram from event camera data.

    Args:
        image_shape: (height, width)
        use_polarity: if True, counts every positive events as +1
                        and every negative events as -1, if False,
                        counts every events as +1
    """
    def __init__(self, image_shape: Tuple[int, int], use_polarity: bool = True) -> None:
        assert image_shape[0] > 0
        assert image_shape[1] > 0
        self.image_shape = image_shape
        self.use_polarity = use_polarity

    def _make_histogram(self, events: np.ndarray) -> np.ndarray:
        y = events[:, 0].astype(int)
        x = events[:, 1].astype(int)
        histogram, _, _ = np.histogram2d(
            y,
            x,
            bins=[self.image_shape[0], self.image_shape[1]],
            range=[[0, self.image_shape[0]], [0, self.image_shape[1]]],
        )
        return histogram  # type: ignore

    def __call__(self, events: np.ndarray) -> np.ndarray:
        """Create histogram.

        Args:
            events: a NumPy array of size [n x d], where n is the number of events and d = 4.
                    Every event is encoded with 4 values (y, x, t, p).

        Returns:
             A 2D histogram where each pixel holds the number of events in it.
        """
        assert events.shape[1] == 4
        if self.use_polarity:
            p = events[:, 3].astype(int)
            pos_events = events[(p == 1)]
            histogram = self._make_histogram(pos_events)
            neg_events = events[(p == 0)]
            histogram -= self._make_histogram(neg_events)
        else:
            histogram = self._make_histogram(events)
        return histogram



from typing import Optional, Tuple, Callable

import torch
import torch.nn as nn
import torch.nn.functional as f
from torch.nn import init


class ConvLayer(nn.Module):
    def __init__(self, in_channels: int, out_channels: int, kernel_size: int, stride:int=1,
                 padding:int=0, activation:Optional[str]='relu', norm:Optional[str]=None) -> None:
        super().__init__()

        bias = False if norm == 'BN' else True
        self.conv2d = nn.Conv2d(in_channels, out_channels, kernel_size, stride, padding, bias=bias)
        if activation is not None:
            self.activation: Callable = getattr(torch, activation, 'relu')  # type: ignore
        else:
            self.activation = None  # type: ignore

        self.norm = norm
        if norm == 'BN':
            self.norm_layer = nn.BatchNorm2d(out_channels)
        elif norm == 'IN':
            self.norm_layer = nn.InstanceNorm2d(out_channels, track_running_stats=True)  # type: ignore

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        out = self.conv2d(x)

        if self.norm in ['BN', 'IN']:
            out = self.norm_layer(out)

        if self.activation is not None:
            out = self.activation(out)

        return out  # type: ignore


class TransposedConvLayer(nn.Module):
    def __init__(self, in_channels: int, out_channels: int, kernel_size: int, stride: int=1,
                 padding: int=0, activation: Optional[str]='relu', norm: Optional[str]=None) -> None:
        super().__init__()

        bias = False if norm == 'BN' else True
        self.transposed_conv2d = nn.ConvTranspose2d(
            in_channels, out_channels, kernel_size, stride=2, padding=padding, output_padding=1, bias=bias)

        if activation is not None:
            self.activation: Callable = getattr(torch, activation, 'relu')  # type: ignore
        else:
            self.activation = None  # type: ignore

        self.norm = norm
        if norm == 'BN':
            self.norm_layer = nn.BatchNorm2d(out_channels)
        elif norm == 'IN':
            self.norm_layer = nn.InstanceNorm2d(out_channels, track_running_stats=True)  # type: ignore

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        out = self.transposed_conv2d(x)
        if self.norm in ['BN', 'IN']:
            out = self.norm_layer(out)
        if self.activation is not None:
            out = self.activation(out)

        return out  # type: ignore


class UpsampleConvLayer(nn.Module):
    def __init__(self, in_channels: int, out_channels: int, kernel_size: int, stride: int=1,
                 padding: int=0, activation: Optional[str]='relu', norm: Optional[str]=None) -> None:
        super().__init__()

        bias = False if norm == 'BN' else True
        self.conv2d = nn.Conv2d(in_channels, out_channels, kernel_size, stride, padding, bias=bias)

        if activation is not None:
            self.activation: Callable = getattr(torch, activation, 'relu')  # type: ignore
        else:
            self.activation = None  # type: ignore

        self.norm = norm
        if norm == 'BN':
            self.norm_layer = nn.BatchNorm2d(out_channels)
        elif norm == 'IN':
            self.norm_layer = nn.InstanceNorm2d(out_channels, track_running_stats=True)   # type: ignore

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x_upsampled = f.interpolate(x, scale_factor=2, mode='bilinear', align_corners=False)
        out = self.conv2d(x_upsampled)
        if self.norm in ['BN', 'IN']:
            out = self.norm_layer(out)
        if self.activation is not None:
            out = self.activation(out)

        return out  # type: ignore


class RecurrentConvLayer(nn.Module):
    def __init__(self, in_channels: int, out_channels: int, kernel_size: int=3,
                 stride: int=1, padding: int=0, recurrent_block_type: str='convlstm',
                 activation: Optional[str]='relu', norm: Optional[str]=None) -> None:
        super().__init__()

        assert(recurrent_block_type in ['convlstm', 'convgru'])
        self.recurrent_block_type = recurrent_block_type
        if self.recurrent_block_type == 'convlstm':
            RecurrentBlock = ConvLSTM
        else:
            raise NotImplementedError("convgru is not supported!")
        self.conv = ConvLayer(in_channels, out_channels, kernel_size, stride, padding, activation, norm)
        self.recurrent_block = RecurrentBlock(input_size=out_channels, hidden_size=out_channels, kernel_size=3)

    def forward(self, x: torch.Tensor, prev_state: Optional[torch.Tensor]) -> Tuple[torch.Tensor,
                                                                                    torch.Tensor]:
        x = self.conv(x)
        state = self.recurrent_block(x, prev_state)
        x = state[0] if self.recurrent_block_type == 'convlstm' else state
        return x, state


# Residual block
class ResidualBlock(nn.Module):
    def __init__(self, in_channels: int, out_channels: int, stride: int=1,
                 downsample: Optional[Callable]=None, norm: Optional[str]=None) -> None:  # TODO: typehint downsample
        super().__init__()
        bias = False if norm == 'BN' else True
        self.conv1 = nn.Conv2d(in_channels, out_channels, kernel_size=3, stride=stride, padding=1, bias=bias)
        self.norm = norm
        if norm == 'BN':
            self.bn1 = nn.BatchNorm2d(out_channels)
            self.bn2 = nn.BatchNorm2d(out_channels)
        elif norm == 'IN':
            self.bn1 = nn.InstanceNorm2d(out_channels)   # type: ignore
            self.bn2 = nn.InstanceNorm2d(out_channels)   # type: ignore

        self.relu = nn.ReLU(inplace=True)
        self.conv2 = nn.Conv2d(out_channels, out_channels, kernel_size=3, stride=1, padding=1, bias=bias)
        self.downsample = downsample

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        residual = x
        out = self.conv1(x)
        if self.norm in ['BN', 'IN']:
            out = self.bn1(out)
        out = self.relu(out)
        out = self.conv2(out)
        if self.norm in ['BN', 'IN']:
            out = self.bn2(out)

        if self.downsample:
            residual = self.downsample(x)

        out += residual
        out = self.relu(out)
        return out   # type: ignore


class ConvLSTM(nn.Module):
    """Adapted from: https://github.com/Atcold/pytorch-CortexNet/blob/master/model/ConvLSTMCell.py """

    def __init__(self, input_size: int, hidden_size: int, kernel_size: int) -> None:
        super().__init__()

        self.input_size = input_size
        self.hidden_size = hidden_size
        pad = kernel_size // 2

        # cache a tensor filled with zeros to avoid reallocating memory at each inference step if --no-recurrent is enabled
        self.zero_tensors = {}  # type: ignore

        self.Gates = nn.Conv2d(input_size + hidden_size, 4 * hidden_size, kernel_size, padding=pad)

    def forward(self, input_: torch.Tensor, prev_state: Optional[torch.Tensor]=None) -> Tuple[torch.Tensor,
                                                                                              torch.Tensor]:

        # get batch and spatial sizes
        batch_size = input_.data.size()[0]
        spatial_size = input_.data.size()[2:]

        # generate empty prev_state, if None is provided
        if prev_state is None:

            # create the zero tensor if it has not been created already
            state_size = tuple([batch_size, self.hidden_size] + list(spatial_size))
            if state_size not in self.zero_tensors:
                # allocate a tensor with size `spatial_size`, filled with zero (if it has not been allocated already)
                self.zero_tensors[state_size] = (
                    torch.zeros(state_size).to(input_.device),
                    torch.zeros(state_size).to(input_.device)
                )

            prev_state = self.zero_tensors[tuple(state_size)]

        prev_hidden, prev_cell = prev_state

        # data size is [batch, channel, height, width]
        stacked_inputs = torch.cat((input_, prev_hidden), 1)
        gates = self.Gates(stacked_inputs)

        # chunk across channel dimension
        in_gate, remember_gate, out_gate, cell_gate = gates.chunk(4, 1)

        # apply sigmoid non linearity
        in_gate = torch.sigmoid(in_gate)
        remember_gate = torch.sigmoid(remember_gate)
        out_gate = torch.sigmoid(out_gate)

        # apply tanh non linearity
        cell_gate = torch.tanh(cell_gate)

        # compute current cell and hidden state
        cell = (remember_gate * prev_cell) + (in_gate * cell_gate)
        hidden = out_gate * torch.tanh(cell)

        return hidden, cell



from typing import Any, Union

import torch
import numpy as np

from .model.model import *
from .utils.inference_utils import CropParameters, EventPreprocessor, IntensityRescaler, ImageFilter, UnsharpMaskFilter
from .utils.loading_utils import get_device

class ImageReconstructor:
    def __init__(self, model: torch.nn.Module, height: int, width: int,
                 num_bins: int, options: Any) -> None:
        self.model = model
        self.use_gpu = options.use_gpu
        self.device = get_device(self.use_gpu)
        self.height = height
        self.width = width
        self.num_bins = num_bins

        self.initialize(self.height, self.width, options)

    def initialize(self, height: int, width: int, options: Any) -> None:
        print('== Image reconstruction == ')
        print('Image size: {}x{}'.format(self.height, self.width))

        self.no_recurrent = options.no_recurrent
        if self.no_recurrent:
            print('!!Recurrent connection disabled!!')

        n_encoders: int = self.model.num_encoders  # type: ignore
        self.crop = CropParameters(self.width, self.height, n_encoders)

        self.last_states_for_each_channel = {'grayscale': None}
        self.event_preprocessor = EventPreprocessor(options)
        self.intensity_rescaler = IntensityRescaler(options)
        self.image_filter = ImageFilter(options)
        self.unsharp_mask_filter = UnsharpMaskFilter(options, device=self.device)

    def update_reconstruction(self, event_tensor: Union[torch.Tensor, np.ndarray]) -> np.ndarray:
        with torch.no_grad():
            if isinstance(event_tensor, np.ndarray):
                event_tensor = torch.from_numpy(event_tensor)

            events = event_tensor.unsqueeze(dim=0)
            events = events.to(self.device)

            events = self.event_preprocessor(events)

            # Resize tensor to [1 x C x crop_size x crop_size] by applying zero padding
            events_for_each_channel = {'grayscale': self.crop.pad(events)}
            reconstructions_for_each_channel = {}
            # Reconstruct new intensity image for each channel (grayscale + RGBW if color reconstruction is enabled)
            for channel in events_for_each_channel.keys():
                new_predicted_frame, states = self.model(events_for_each_channel[channel],
                                                            self.last_states_for_each_channel[channel])

                if self.no_recurrent:
                    self.last_states_for_each_channel[channel] = None
                else:
                    self.last_states_for_each_channel[channel] = states

                # Output reconstructed image
                crop = self.crop if channel == 'grayscale' else self.crop_halfres  # type: ignore

                # Unsharp mask (on GPU)
                new_predicted_frame = self.unsharp_mask_filter(new_predicted_frame)
                # Intensity rescaler (on GPU)
                new_predicted_frame = self.intensity_rescaler(new_predicted_frame)
                reconstructions_for_each_channel[channel] = new_predicted_frame[0, 0, crop.iy0:crop.iy1,
                                                                                crop.ix0:crop.ix1].cpu().numpy()

                out = reconstructions_for_each_channel['grayscale']

            # Post-processing, e.g bilateral filter (on CPU)
            out = self.image_filter(out)
        return out  # type: ignore



from typing import Optional, List, Tuple, Callable

import torch
import torch.nn as nn
from .submodules import ConvLayer, UpsampleConvLayer, TransposedConvLayer, RecurrentConvLayer, ResidualBlock


def skip_concat(x1: torch.Tensor, x2: torch.Tensor) -> torch.Tensor:
    return torch.cat([x1, x2], dim=1)


def skip_sum(x1: torch.Tensor, x2: torch.Tensor) -> torch.Tensor:
    return x1 + x2


class BaseUNet(nn.Module):
    def __init__(self, num_input_channels: int, num_output_channels: int=1,
                 skip_type: str='sum', activation: str='sigmoid', num_encoders: int=4,
                 base_num_channels: int=32, num_residual_blocks: int=2, norm: Optional[str]=None,
                 use_upsample_conv: bool=True) -> None:
        super(BaseUNet, self).__init__()

        self.num_input_channels = num_input_channels
        self.num_output_channels = num_output_channels
        self.skip_type = skip_type
        self.apply_skip_connection = skip_sum if self.skip_type == 'sum' else skip_concat
        self.activation_func = activation
        self.norm = norm

        if use_upsample_conv:
            print('Using UpsampleConvLayer (slow, but no checkerboard artefacts)')
            self.UpsampleLayer = UpsampleConvLayer
        else:
            print('Using TransposedConvLayer (fast, with checkerboard artefacts)')
            self.UpsampleLayer = TransposedConvLayer  # type: ignore

        self.num_encoders = num_encoders
        self.base_num_channels = base_num_channels
        self.num_residual_blocks = num_residual_blocks
        self.max_num_channels = self.base_num_channels * pow(2, self.num_encoders)

        assert(self.num_input_channels > 0)
        assert(self.num_output_channels > 0)

        self.encoder_input_sizes = []
        for i in range(self.num_encoders):
            self.encoder_input_sizes.append(self.base_num_channels * pow(2, i))

        self.encoder_output_sizes = [self.base_num_channels * pow(2, i + 1) for i in range(self.num_encoders)]

        self.activation: Callable = getattr(torch, self.activation_func, 'sigmoid')  # type: ignore

    def build_resblocks(self) -> None:
        self.resblocks = nn.ModuleList()
        for i in range(self.num_residual_blocks):
            self.resblocks.append(ResidualBlock(self.max_num_channels, self.max_num_channels, norm=self.norm))

    def build_decoders(self) -> None:
        decoder_input_sizes = list(reversed([self.base_num_channels * pow(2, i + 1) for i in range(self.num_encoders)]))

        self.decoders = nn.ModuleList()
        for input_size in decoder_input_sizes:
            self.decoders.append(self.UpsampleLayer(input_size if self.skip_type == 'sum' else 2 * input_size,
                                                    input_size // 2,
                                                    kernel_size=5, padding=2, norm=self.norm))

    def build_prediction_layer(self) -> None:
        self.pred = ConvLayer(self.base_num_channels if self.skip_type == 'sum' else 2 * self.base_num_channels,
                              self.num_output_channels, 1, activation=None, norm=self.norm)


class UNet(BaseUNet):
    def __init__(self, num_input_channels: int, num_output_channels: int=1,
                 skip_type: str='sum', activation: Optional[str]='sigmoid', num_encoders: int=4,
                 base_num_channels: int=32, num_residual_blocks: int=2, norm: Optional[str]=None,
                 use_upsample_conv: bool=True) -> None:
        super().__init__(num_input_channels, num_output_channels, skip_type, activation,   # type: ignore
                         num_encoders, base_num_channels, num_residual_blocks, norm, use_upsample_conv)

        self.head = ConvLayer(self.num_input_channels, self.base_num_channels,
                              kernel_size=5, stride=1, padding=2)  # N x C x H x W -> N x 32 x H x W

        self.encoders = nn.ModuleList()
        for input_size, output_size in zip(self.encoder_input_sizes, self.encoder_output_sizes):
            self.encoders.append(ConvLayer(input_size, output_size, kernel_size=5,
                                           stride=2, padding=2, norm=self.norm))

        self.build_resblocks()
        self.build_decoders()
        self.build_prediction_layer()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        :param x: N x num_input_channels x H x W
        :return: N x num_output_channels x H x W
        """

        # head
        x = self.head(x)
        head = x

        # encoder
        blocks = []
        for i, encoder in enumerate(self.encoders):
            x = encoder(x)
            blocks.append(x)

        # residual blocks
        for resblock in self.resblocks:
            x = resblock(x)

        # decoder
        for i, decoder in enumerate(self.decoders):
            x = decoder(self.apply_skip_connection(x, blocks[self.num_encoders - i - 1]))

        img = self.activation(self.pred(self.apply_skip_connection(x, head)))
        return img  # type: ignore


class UNetRecurrent(BaseUNet):
    """
    Recurrent UNet architecture where every encoder is followed by a recurrent convolutional block,
    such as a ConvLSTM or a ConvGRU.
    Symmetric, skip connections on every encoding layer.
    """

    def __init__(self, num_input_channels: int, num_output_channels: int=1, skip_type: str='sum',
                 recurrent_block_type: str='convlstm', activation: Optional[str]='sigmoid',
                 num_encoders: int=4, base_num_channels: int=32, num_residual_blocks: int=2,
                 norm: Optional[str]=None, use_upsample_conv: bool=True) -> None:
        super().__init__(num_input_channels, num_output_channels, skip_type, activation,   # type: ignore
                         num_encoders, base_num_channels, num_residual_blocks, norm,
                         use_upsample_conv)

        self.head = ConvLayer(self.num_input_channels, self.base_num_channels,
                              kernel_size=5, stride=1, padding=2)  # N x C x H x W -> N x 32 x H x W

        self.encoders = nn.ModuleList()
        for input_size, output_size in zip(self.encoder_input_sizes, self.encoder_output_sizes):
            self.encoders.append(RecurrentConvLayer(input_size, output_size,
                                                    kernel_size=5, stride=2, padding=2,
                                                    recurrent_block_type=recurrent_block_type,
                                                    norm=self.norm))

        self.build_resblocks()
        self.build_decoders()
        self.build_prediction_layer()

    def forward(self, x: torch.Tensor, prev_states: Optional[List[torch.Tensor]]) -> Tuple[torch.Tensor,
                                                                                           List[torch.Tensor]]:
        """
        :param x: N x num_input_channels x H x W
        :param prev_states: previous LSTM states for every encoder layer
        :return: N x num_output_channels x H x W
        """

        # head
        x = self.head(x)
        head = x

        if prev_states is None:
            prev_states = [None] * self.num_encoders  # type: ignore

        # encoder
        blocks = []
        states = []
        for i, encoder in enumerate(self.encoders):
            x, state = encoder(x, prev_states[i])
            blocks.append(x)
            states.append(state)

        # residual blocks
        for resblock in self.resblocks:
            x = resblock(x)

        # decoder
        for i, decoder in enumerate(self.decoders):
            x = decoder(self.apply_skip_connection(x, blocks[self.num_encoders - i - 1]))

        # tail
        img = self.activation(self.pred(self.apply_skip_connection(x, head)))

        return img, states



from typing import Any
import logging
import torch.nn as nn
import numpy as np


class BaseModel(nn.Module):
    """
    Base class for all models
    """
    def __init__(self, config) -> None:  # type: ignore
        super().__init__()
        self.config = config
        self.logger = logging.getLogger(self.__class__.__name__)

    def forward(self, *input):  # type: ignore
        """
        Forward pass logic

        :return: Model output
        """
        raise NotImplementedError

    def summary(self) -> None:
        """
        Model summary
        """
        model_parameters = filter(lambda p: p.requires_grad, self.parameters())
        params = sum([np.prod(p.size()) for p in model_parameters])
        self.logger.info('Trainable parameters: {}'.format(params))
        self.logger.info(self)



from typing import Any
from math import ceil, floor
from collections import deque

from torch.nn import ReflectionPad2d
import numpy as np
import torch
import cv2
import scipy.stats as st
import torch.nn.functional as F


def gkern(kernlen: int=5, nsig: float=1.0) -> torch.Tensor:
    """Returns a 2D Gaussian kernel array."""
    """https://stackoverflow.com/a/29731818"""
    interval = (2 * nsig + 1.) / (kernlen)
    x = np.linspace(-nsig - interval / 2., nsig + interval / 2., kernlen + 1)
    kern1d = np.diff(st.norm.cdf(x))
    kernel_raw = np.sqrt(np.outer(kern1d, kern1d))
    kernel = kernel_raw / kernel_raw.sum()
    return torch.from_numpy(kernel).float()


class EventPreprocessor:
    """
    Utility class to preprocess event tensors.
    Can perform operations such as hot pixel removing, event tensor normalization,
    or flipping the event tensor.
    """

    def __init__(self, options: Any) -> None:

        print('== Event preprocessing ==')
        self.no_normalize = options.no_normalize
        if self.no_normalize:
            print('!!Will not normalize event tensors!!')
        else:
            print('Will normalize event tensors.')

        self.flip = options.flip
        if self.flip:
            print('Will flip event tensors.')

    def __call__(self, events: torch.Tensor) -> torch.Tensor:
        # Flip tensor vertically and horizontally
        if self.flip:
            events = torch.flip(events, dims=[2, 3])

        # Normalize the event tensor (voxel grid) so that
        # the mean and stddev of the nonzero values in the tensor are equal to (0.0, 1.0)
        if not self.no_normalize:
            nonzero_ev = (events != 0)
            num_nonzeros = nonzero_ev.sum()
            if num_nonzeros > 0:
                # compute mean and stddev of the **nonzero** elements of the event tensor
                # we do not use PyTorch's default mean() and std() functions since it's faster
                # to compute it by hand than applying those funcs to a masked array
                mean = events.sum() / num_nonzeros
                stddev = torch.sqrt((events ** 2).sum() / num_nonzeros - mean ** 2)
                mask = nonzero_ev.float()
                events = mask * (events - mean) / stddev

        return events


class IntensityRescaler:
    """
    Utility class to rescale image intensities to the range [0, 1],
    using (robust) min/max normalization.
    Optionally, the min/max bounds can be smoothed over a sliding window to avoid jitter.
    """

    def __init__(self, options: Any) -> None:
        self.auto_hdr = options.auto_hdr
        self.intensity_bounds = deque()  # type: ignore
        self.auto_hdr_median_filter_size = options.auto_hdr_median_filter_size
        self.Imin = options.Imin
        self.Imax = options.Imax

    def __call__(self, img: torch.Tensor) -> torch.Tensor:
        """
        param img: [1 x 1 x H x W] Tensor taking values in [0, 1]
        """
        if self.auto_hdr:
            Imin = torch.min(img).item()
            Imax = torch.max(img).item()

            # ensure that the range is at least 0.1
            Imin = np.clip(Imin, 0.0, 0.45)
            Imax = np.clip(Imax, 0.55, 1.0)

            # adjust image dynamic range (i.e. its contrast)
            if len(self.intensity_bounds) > self.auto_hdr_median_filter_size:
                self.intensity_bounds.popleft()

            self.intensity_bounds.append((Imin, Imax))
            self.Imin = np.median([rmin for rmin, rmax in self.intensity_bounds])
            self.Imax = np.median([rmax for rmin, rmax in self.intensity_bounds])

        img = 255.0 * (img - self.Imin) / (self.Imax - self.Imin)
        img.clamp_(0.0, 255.0)
        img = img.byte()  # convert to 8-bit tensor

        return img


class UnsharpMaskFilter:
    """
    Utility class to perform unsharp mask filtering on reconstructed images.
    """
    def __init__(self, options: Any, device: torch.device) -> None:
        self.unsharp_mask_amount = options.unsharp_mask_amount
        self.unsharp_mask_sigma = options.unsharp_mask_sigma
        self.gaussian_kernel_size = 5
        self.gaussian_kernel = gkern(self.gaussian_kernel_size,
                                     self.unsharp_mask_sigma).unsqueeze(0).unsqueeze(0).to(device)

    def __call__(self, img: torch.Tensor) -> torch.Tensor:
        if self.unsharp_mask_amount > 0:
            blurred = F.conv2d(img, self.gaussian_kernel,
                                padding=self.gaussian_kernel_size // 2)
            img = (1 + self.unsharp_mask_amount) * img - self.unsharp_mask_amount * blurred
        return img


class ImageFilter:
    """
    Utility class to perform some basic filtering on reconstructed images.
    """

    def __init__(self, options: Any) -> None:
        self.bilateral_filter_sigma = options.bilateral_filter_sigma

    def __call__(self, img: torch.Tensor) -> torch.Tensor:
        if self.bilateral_filter_sigma:
            # TODO check - this looks like returning numpy, not torch.
            filtered_img = np.zeros_like(img)
            filtered_img = cv2.bilateralFilter(
                img, 5, 25.0 * self.bilateral_filter_sigma, 25.0 * self.bilateral_filter_sigma)
            img = filtered_img  # type: ignore
        return img


def optimal_crop_size(max_size: int, max_subsample_factor: int) -> int:
    """ Find the optimal crop size for a given max_size and subsample_factor.
        The optimal crop size is the smallest integer which is greater or equal than max_size,
        while being divisible by 2^max_subsample_factor.
    """
    crop_size = int(pow(2, max_subsample_factor) * ceil(max_size / pow(2, max_subsample_factor)))
    return crop_size


class CropParameters:
    """ Helper class to compute and store useful parameters for pre-processing and post-processing
        of images in and out of E2VID.
        Pre-processing: finding the best image size for the network, and padding the input image with zeros
        Post-processing: Crop the output image back to the original image size
    """

    def __init__(self, width: int, height: int, num_encoders: int) -> None:
        self.height = height
        self.width = width
        self.num_encoders = num_encoders
        self.width_crop_size = optimal_crop_size(self.width, num_encoders)
        self.height_crop_size = optimal_crop_size(self.height, num_encoders)

        self.padding_top = ceil(0.5 * (self.height_crop_size - self.height))
        self.padding_bottom = floor(0.5 * (self.height_crop_size - self.height))
        self.padding_left = ceil(0.5 * (self.width_crop_size - self.width))
        self.padding_right = floor(0.5 * (self.width_crop_size - self.width))
        self.pad = ReflectionPad2d((self.padding_left, self.padding_right, self.padding_top, self.padding_bottom))

        self.cx = floor(self.width_crop_size / 2)
        self.cy = floor(self.height_crop_size / 2)

        self.ix0 = self.cx - floor(self.width / 2)
        self.ix1 = self.cx + ceil(self.width / 2)
        self.iy0 = self.cy - floor(self.height / 2)
        self.iy1 = self.cy + ceil(self.height / 2)



"""Custom data types."""

from typing import Tuple, Union

import numpy as np
import torch

from .raw_event import RawEvent
from .raw_events import RawEvents

NUMPY_TORCH = Union[np.ndarray, torch.Tensor]
FLOAT_TORCH = Union[float, torch.Tensor]



from typing import Any
import logging

import torch
from ..model.model import *

logger = logging.getLogger('ev_lib')


def load_model(path_to_model: str) -> Any:
    print('Loading model {}...'.format(path_to_model))
    raw_model = torch.load(path_to_model, map_location=torch.device('cpu'))  # TODO handle CPU / GPU switch here
    arch = raw_model['arch']

    try:
        model_type = raw_model['model']
    except KeyError:
        model_type = raw_model['config']['model']

    # instantiate model
    model = eval(arch)(model_type)

    # load model weights
    model.load_state_dict(raw_model['state_dict'])
    return model


def get_device(use_gpu: bool) -> torch.device:
    if use_gpu and torch.cuda.is_available():
        device = torch.device('cuda:0')
    else:
        device = torch.device('cpu')
    logger.info(f'Device: {device}')

    return device



"""Text format.
"""
import logging
import os
from typing import Any
from typing import Dict
from typing import List

import cv2
import numpy as np


logger = logging.getLogger(__name__)

from ...types import RawEvents
from ._iterator_access import IteratorAccess


# TODO make parser abstract and merge these classes.


class IteratorText(IteratorAccess):
    FORMAT = "text"

    def __init__(self, textfile: str) -> None:
        # TODO add parse format option
        super().__init__(textfile)
        self.file = open(self.file_name)

    def __iter__(self) -> Any:
        self.count = 0
        return self

    def read_next_lines(self) -> List[str]:
        lines = self.file.readlines()
        if not lines:
            raise StopIteration
        return lines


class IteratorTextEvent(IteratorText):
    def __next__(self) -> RawEvents:
        """
        Returns:
            RawEvents: events
        """
        lines = self.read_next_lines()
        _l = len(lines)
        x = np.zeros((_l,), dtype=np.int32)
        y = np.zeros((_l,), dtype=np.int32)
        t = np.zeros((_l,), dtype=np.float64)
        p = np.zeros((_l,), dtype=bool)
        for _i, line in enumerate(lines):
            val = line.split()
            t[_i] = np.float64(val[0])
            x[_i] = int(val[1])
            y[_i] = int(val[2])
            p[_i] = int(val[3])
        self.count += _l
        return RawEvents(x=x, y=y, timestamp=t, polarity=p)


class IteratorTextPose(IteratorText):
    def __next__(self) -> Dict[str, Any]:
        """
        Returns:
            dict: {"pose": np.ndarray (N, 7)}
        """
        lines = self.read_next_lines()
        _l = len(lines)
        pose_list = []
        for _i, line in enumerate(lines):
            val = line.split()
            pose_list.append(np.array(val, dtype=np.float64))
        self.count += _l
        return {"pose": np.stack(pose_list, axis=0), "num": _l}


class IteratorTextImu(IteratorText):
    def __next__(self) -> Dict[str, Any]:
        """
        Returns:
            dict: {"imu": np.ndarray (N, 7)}
        """
        lines = self.read_next_lines()
        _l = len(lines)
        imu_list = []
        for _i, line in enumerate(lines):
            val = line.split()
            imu_list.append(np.array(val, dtype=np.float64))
        self.count += _l
        return {"imu": np.stack(imu_list, axis=0), "num": _l}


class IteratorTextFrame(IteratorText):
    def __next__(self) -> Dict[str, Any]:
        """
        Returns:
            dict: {"frame": np.ndarray (N, H, W), "t": np.ndarray (N)}
        """
        # TODO assumption is that always images.txt and the image root dir is the same.
        image_dir = os.path.dirname(self.file_name)
        lines = self.read_next_lines()
        _l = len(lines)
        image_list = []
        image_timestamp = []
        for line in lines:
            val = line.split()
            image_timestamp.append(np.float64(val[0]))
            image_file = os.path.join(image_dir, val[1])
            image_list.append(cv2.imread(image_file, cv2.IMREAD_GRAYSCALE))
        self.count += _l
        image_timestamps = np.array(image_timestamp)
        images = np.array(image_list)
        return {"frame": images, "t": image_timestamps, "num": _l}



"""aedat file formats, mainly for DV software by iniVation.
"""
import logging
from typing import Any
from typing import Dict

import dv
import numpy as np
from dv import AedatFile


logger = logging.getLogger(__name__)

from ...types import RawEvents
from ._iterator_access import IteratorAccess


# TODO make parser abstract and merge these classes.


class IteratorAedat4(IteratorAccess):
    FORMAT = "aedat4"

    def __init__(self, aedat4file: str) -> None:
        super().__init__(aedat4file)
        self.file = dv.AedatFile(self.file_name)
        self.count = 0


class IteratorAedat4Event(IteratorAedat4):
    def __init__(self, aedat4file: str) -> None:
        super().__init__(aedat4file)
        self.file_iter = self.file["events"].numpy()

    def __next__(self) -> RawEvents:
        """
        Returns:
            RawEvents:
        """
        ev = self.file_iter.__next__()
        return RawEvents(x=ev["x"], y=ev["y"], timestamp=ev["timestamp"], polarity=ev["polarity"])


class IteratorAedat4Trigger(IteratorAedat4):
    def __init__(self, aedat4file: str) -> None:
        super().__init__(aedat4file)
        self.file_iter = self.file["triggers"]

    def __next__(self) -> Dict[str, Any]:
        """
        Returns:
            dict: {"trigger": (1, 2)}
        """
        tr = self.file_iter.__next__()
        trigger = np.array([[tr.timestamp, tr.type]], dtype=np.float64)
        return {"trigger": trigger, "num": len(trigger)}


class IteratorAedat4Imu(IteratorAedat4):
    def __init__(self, aedat4file: str) -> None:
        super().__init__(aedat4file)
        self.file_iter = self.file["imu"]

    def __next__(self) -> Dict[str, Any]:
        """
        Returns:
            dict: {"imu": np.ndarray (1, N, 7)}
            7 is t, ax, ay, az, gx, gy, gz
        """
        imu = self.file["imu"].__next__()
        return {
            "imu": np.array([imu.timestamp, *imu.accelerometer, *imu.gyroscope]),
            "num": 1,
        }


class IteratorAedat4Frame(IteratorAedat4):
    def __init__(self, aedat4file: str) -> None:
        super().__init__(aedat4file)
        self.file_iter = self.file["frames"]

    def __next__(self) -> Dict[str, Any]:
        """
        Returns:
            dict: {"t" (1, ), "image": (1, H, W, 3)}
        """
        fr = self.file_iter.__next__()
        ts = np.array([fr.timestamp])
        image = np.array(fr.image, dtype=np.uint8)[None]  # H, W, 1
        return {"frame": image, "t": ts}



"""Basic utility functions."""
import numpy as np

def generate_events(
    n_events: int,
    height: int,
    width: int,
    tmin: float = 0.0,
    tmax: float = 0.5,
    dist: str = "uniform",
) -> np.ndarray:
    """Generate random events.

    Args:
        n_events (int): num of events
        height (int): height of the camera
        width (int): width of the camera
        tmin (float): timestamp min
        tmax (float): timestamp max
        dist (str): distribution of generated events. currently only "uniform" is supported.

    Returns:
        events (np.ndarray): [n_events x 4] numpy array.
            (y (height), x (width), t: [tmin, tmax], p: {0, 1})
    """
    if dist != "uniform":
        raise NotImplementedError
    y = np.random.randint(0, height, n_events)
    x = np.random.randint(0, width, n_events)
    t = np.random.uniform(tmin, tmax, n_events)
    t = np.sort(t)
    p = np.random.randint(0, 2, n_events)
    events: np.ndarray = np.concatenate([y[..., None], x[..., None], t[..., None], p[..., None]], axis=1)
    return events


def generate_random_optical_flow(image_size: tuple, max_val: float) -> np.ndarray:
    """Generate random optical flow.

    Args:
        image_size (tuple): (H, W)
        max_val (float): max value 

    Returns:
        flow (np.ndarray): [2 x H x W] array.
    """
    flow = np.random.uniform(-max_val, max_val, (2,) + image_size)
    return flow


def generate_uniform_optical_flow(image_size: tuple, vertical: float, horizontal: float) -> np.ndarray:
    """Generate uniform optical flow.

    Args:
        image_size (tuple): (H, W)
        vertical (float): vertical direction component
        horizontal (float): horizontal direction component

    Returns:
        flow (np.ndarray): [2 x H x W] array.
    """
    flow = np.ones((2,) + image_size) * np.array([vertical, horizontal])[:, None, None]
    return flow # type: ignore



from typing import Any
from typing import Dict


class IteratorAccess:
    FORMAT = "base"

    def __init__(self, file_name: str, **kwargs: Any) -> None:
        self.file_name = file_name

    def __iter__(self) -> Any:
        return self

    def __next__(self) -> Any:
        raise NotImplementedError



"""Time-map representation.
"""
from typing import Any
from typing import Tuple

import numpy as np


class TimeMap:
    """Create a time map of events. Adapted from:
    Lagorce, Xavier, et al. "Hots: a hierarchy of event-based time-surfaces for pattern recognition."
    IEEE transactions on pattern analysis and machine intelligence 39.7 (2016): 1346-1359.

    Note that this implementation is a "random access implementation" and does not hold a state.

    Args:
        image_shape: (height, width)
        decay: the factor in the exponential. A higher value leads to a stronger decay (sharper edges).
    """
    def __init__(self, image_shape: Tuple[int, int], decay: float) -> None:
        assert image_shape[0] > 0
        assert image_shape[1] > 0
        self.image_shape = image_shape
        self.decay = decay

    def __call__(self, events: np.ndarray) -> np.ndarray:
        """Create time map.

        Args:
            events: a NumPy array of size [n x d], where n is the number of events and d = 4.
                    Every event is encoded with 4 values (y, x, t, p).

        Returns:
            A 2D representation of the input events
        """
        assert events.shape[1] == 4
        time_map = np.zeros(self.image_shape)
        y = events[:, 0].astype(int)
        x = events[:, 1].astype(int)
        t = events[:, 2]
        time_map[y, x] = t
        return np.exp(-self.decay * (np.amax(t) - time_map))  # type: ignore



"""2-D visualization of various quantities.
"""
import glob
import logging
import os
from typing import Any, Dict, List, Optional, Tuple

import cv2
import numpy as np

logger = logging.getLogger(__name__)


from PIL import Image, ImageDraw

from . import utils

TRANSPARENCY = 0.25  # Degree of transparency, 0-100%
OPACITY = int(255 * TRANSPARENCY)

MIN_DEPTH = 0.1
MAX_DEPTH = 100


def image(image: Any) -> Image.Image:
    """Visualize image.

    Args:
        image (Any): str, np.ndarray, or PIL Image.

    Returns:
        Image.Image: PIL Image object
    """
    image = utils.load_image(image)
    return image


def depth(depth: np.ndarray) -> Image.Image:
    """Visualize depth image.

    Args:
        depth (np.ndarray): depth image.

    Returns:
        Image.Image: PIL Image object, depth
    """
    image = utils.color_depth_with_nan(depth, MIN_DEPTH, MAX_DEPTH, MAX_DEPTH)
    image = Image.fromarray(image)
    return image


def optical_flow(
    flow_v: np.ndarray,
    flow_h: np.ndarray,
    visualize_color_wheel: bool = True,
    ord: float = 0.5,
) -> Tuple[np.ndarray, Optional[np.ndarray]]:
    """Visualize optical flow.
    
    Args:
        flow_v (numpy.ndarray): [H x W], vertical direction.
        flow_h (numpy.ndarray): [H x W], horizontal direction.
        visualize_color_wheel (bool): If True, it also visualizes the color wheel (legend for OF).
        file_prefix (Optional[str], optional): [description]. Defaults to None.
            If specified, the save location will be `save_dir/{prefix}_{unique}.png`.

    Returns:
        image (PIL.Image): PIL image.
    """
    flow_rgb, color_wheel, _ = utils.color_optical_flow(flow_v, flow_h, ord=ord)
    image = Image.fromarray(flow_rgb)

    if visualize_color_wheel:
        wheel = Image.fromarray(color_wheel)
    else:
        wheel = None
    return image, wheel



"""Utility functions for coloring, loading, etc.
"""
import logging
import os
from typing import Any, Dict, List, Optional, Tuple

import cv2
import numpy as np
from matplotlib import pyplot as plt


from typing import Any, Dict, List, Optional
from PIL import Image


def load_image(image: Any) -> Image.Image:
    """A wrapper function to get image and returns PIL Image object.

    Args:
        image (str or np.ndarray): If it is str, open and load the image.
            If it is numpy array, it converts to PIL.Image.

    Returns:
        Image.Image: PIl Image object.
    """
    if type(image) == str:
        image = Image.open(image)
    elif type(image) == np.ndarray:
        image = Image.fromarray(image)
    return image


def color_depth_with_nan(
    depth: np.ndarray, min_depth: float, max_depth: float, fillnan: float
) -> np.ndarray:
    """Color depth image even with nan values. The nan will be padded with `fillnan` value.

    Args:
        depth (np.ndarray): [1, H, W]?

    Returns:
        [H, W, 3] colored depth arrat.
    """
    depth[np.isnan(depth)] = fillnan
    return color_depth(depth, min_depth, max_depth)


def color_depth(depth: np.ndarray, min_depth: float, max_depth: float) -> np.ndarray:
    """Color depth image.
    
    Args:
        depth (np.ndarray): [1, H, W]?  # TODO test and check

    Returns:
        [H, W, 3] colored depth array.
    """
    depth = np.clip(depth, min_depth, max_depth)
    depth_relative = (depth - min_depth) / (max_depth - min_depth)
    color_depth = (255 * plt.cm.viridis(depth_relative.astype(np.float32))[0, :, :, :3]).astype(np.uint8)
    return color_depth  # type: ignore


def color_optical_flow(
    flow_v: np.ndarray,
    flow_h: np.ndarray,
    max_magnitude: Optional[float]=None,
    ord:float=1.0
) -> Tuple[np.ndarray, np.ndarray, float]:
    """Color optical flow.
    
    Args:
        flow_v (numpy.ndarray): [H x W], vertical direction.
        flow_h (numpy.ndarray): [H x W], horizontal direction.
        max_magnitude (float, optional): Max magnitude used for the colorization. Defaults to None.
        ord (float): 1: our usual, 0.5: DSEC colorinzing.

    Returns:
        flow_rgb (np.ndarray): [W, H, 3]
        color_wheel (np.ndarray): [H, H, 3] color wheel
        max_magnitude (float): max magnitude of the flow.
    """
    imshape = flow_h.shape
    flows = np.stack((flow_v, flow_h), axis=2)
    flows[np.isinf(flows)] = 0
    flows[np.isnan(flows)] = 0
    mag = np.linalg.norm(flows, axis=2) ** ord
    ang = (np.arctan2(flow_h, flow_v) + np.pi) * 180.0 / np.pi / 2.0
    ang = ang.astype(np.uint8)
    hsv = np.zeros([imshape[0], imshape[1], 3], dtype=np.uint8)
    hsv[:, :, 0] = ang
    hsv[:, :, 1] = 255
    if max_magnitude is None:
        max_magnitude = mag.max()
    hsv[:, :, 2] = (255 * mag / max_magnitude).astype(np.uint8)
    # hsv[:, :, 2] = cv2.normalize(mag, None, 0, 255, cv2.NORM_MINMAX)
    flow_rgb = cv2.cvtColor(hsv, cv2.COLOR_HSV2RGB)

    # Color wheel
    hsv = np.zeros([imshape[0], imshape[0], 3], dtype=np.uint8)  # square
    xx, yy = np.meshgrid(
        np.linspace(-1, 1, imshape[0]), np.linspace(-1, 1, imshape[0])
    )
    mag = np.linalg.norm(np.stack((xx, yy), axis=2), axis=2)
    # ang = (np.arctan2(yy, xx) + np.pi) * 180 / np.pi / 2.0
    ang = (np.arctan2(xx, yy) + np.pi) * 180 / np.pi / 2.0
    hsv[:, :, 0] = ang.astype(np.uint8)
    hsv[:, :, 1] = 255
    hsv[:, :, 2] = (255 * mag / mag.max()).astype(np.uint8)
    # hsv[:, :, 2] = cv2.normalize(mag, None, 0, 255, cv2.NORM_MINMAX)
    color_wheel = cv2.cvtColor(hsv, cv2.COLOR_HSV2RGB)

    return flow_rgb, color_wheel, max_magnitude



"""evk file formats, mainly for Metavision SDK software by Prophesee.
"""
import logging
from typing import Any
from typing import Dict

import numpy as np
from expelliarmus import Wizard


logger = logging.getLogger(__name__)

from ._iterator_access import IteratorAccess


# TODO make parser abstract and merge these classes.


class IteratorEvk3(IteratorAccess):
    FORMAT = "evk3"

    def __init__(self, evk3file: str) -> None:
        super().__init__(evk3file)
        self.wizard = Wizard(encoding="evt3", fpath=self.file_name, chunk_size=16384)
        self.count = 0

    def __next__(self) -> Dict[str, Any]:
        return self.wizard.read_chunk()  # type: ignore



"""Hot-pixel removal filters.
"""
import numpy as np
import torch

from ..types import NUMPY_TORCH



def hot_pixel_filter(events: NUMPY_TORCH, image_shape: tuple, hot_pixel: int = 10) -> NUMPY_TORCH:
    """Apply hot-pixel removal, a speed-optimized implementation.
        The filtering is for a single batch of events, not multiple or asynchronous.

    Args:
        events (NUMPY_TORCH): [N, 4]
        image_shape (tuple): tuple of (H, W).
        hot_pixel (int): threshold of hot pixels.

    Returns:
        NUMPY_TORCH: Filered events
    """
    xy = events[..., :2]
    if isinstance(events, np.ndarray):
        pix_index = (xy[..., 0] + xy[..., 1] * image_shape[0]).astype(np.int64)   # unique indices at all pixels
        val, counts = np.unique(pix_index, return_counts=True)
        dup_vals = val[counts > hot_pixel]  # the unique index for hot pixel locations
        hot_event_index = np.in1d(pix_index, dup_vals)  # bool of the events belong to the hot pixels
    elif isinstance(events, torch.Tensor):
        pix_index = (xy[..., 0] + xy[..., 1] * image_shape[0]).long()   # unique indices at all pixels
        val, counts = torch.unique(pix_index, return_counts=True)
        dup_vals = val[counts > hot_pixel]  # the unique index for hot pixel locations
        hot_event_index = torch.isin(pix_index, dup_vals)  # bool of the events belong to the hot pixels
    filtered_event = events[~hot_event_index]
    return filtered_event




"""To be used: define dataset.
"""

from typing import Any


class IteratorAccessDataset:
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        pass


def setup(*args: Any, **kwargs: Any) -> IteratorAccessDataset:
    return IteratorAccessDataset(args, kwargs)



"""Data type for a single event."""
import copy
from dataclasses import dataclass
from typing import Any, List, Optional

import numpy as np
import numpy.typing as npt


@dataclass
class RawEvent:
    """Dataclass for raw events"""

    x: np.int16  # [0, width]
    y: np.int16  # [0, height]
    timestamp: np.float64
    polarity: bool  # true for positive, false for negative

    # Shortcuts
    @property
    def p(self) -> bool:
        """Alias for polatiry.
        """
        return self.polarity

    @property
    def t(self) -> np.float64:
        """Alias for timestamp.
        """
        return self.timestamp

    @property
    def color(self) -> tuple:
        if self.p:
            return (255, 0, 0)  # Red
        else:
            return (0, 0, 255)  # Blue

    def copy(self) -> Any:
        return copy.deepcopy(self)

    def as_numpy(self) -> npt.NDArray[np.float64]:
        """Convert event object into 1-d numpy array.

        Returns:
            npt.NDArray[np.float64]: 1-d numpy array.
        """
        return np.array([self.y, self.x, self.t, self.p], dtype=np.float64)
